<?php

namespace LakeDrops\DockerTraefik;

use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

/**
 * Class Traefik.
 *
 * @package LakeDrops\DockerTraefik
 */
class Traefik {

  /**
   * The name of the project using Traefik.
   *
   * @var string
   */
  protected string $name;

  /**
   * @var string
   */
  protected string $domain;

  /**
   * @var int
   */
  protected int $http_port;

  /**
   * @var int
   */
  protected int $https_port;

  /**
   * @var string
   */
  protected string $cert_filename;

  /**
   * @var string
   */
  protected string $key_filename;

  /**
   * @var bool
   */
  protected bool $addon_portainer = FALSE;

  /**
   * @var string
   */
  protected string $hub_token = '';

  /**
   * @var array
   */
  protected array $env;

  /**
   * @var bool
   */
  protected bool $tls;

  /**
   * @var bool
   */
  protected bool $dns_challenge;

  /**
   * @var string
   */
  protected string $dns_challenge_provider;

  /**
   * @var string
   */
  protected string $dns_challenge_resolver;

  /**
   * Traefik constructor.
   *
   * @param string $name
   *   Name of the network to be created which needs to match the container
   *   prefix of your project you would like to handle with Traefik.
   * @param string $domain
   *   The domain name used for all local projects.
   * @param int $http_port
   *   Port for non secure requests.
   * @param int $https_port
   *   Port for secure requests.
   * @param string $cert_filename
   *   Filename of the SSL certificate.
   * @param string $key_filename
   *   Filename of the private key for the SSL certificate.
   * @param array $env
   *   A list of environment variables for the Traefik container.
   * @param bool $tls
   *   Whether TLS should be supported.
   * @param bool $dns_challenge
   *   If TLS is supported, it uses the http challenge by default. Set to TRUE
   *   to use the DNS challenge.
   * @param string $dns_challenge_provider
   *   If DNS challenge should be used, a provider is required. For a list of
   *   supported providers:
   *   @see https://doc.traefik.io/traefik/https/acme/#providers
   * @param string $dns_challenge_resolver
   *   In some scenarios, the DNS resolver needs to be defined in order to
   *   prevent local or shadow DNS servers being used.
   */
  public function __construct(
    string $name,
    string $domain = 'docker.localhost',
    int $http_port = 8000,
    int $https_port = 8443,
    string $cert_filename = '',
    string $key_filename = '',
    array $env = [],
    bool $tls = FALSE,
    bool $dns_challenge = FALSE,
    string $dns_challenge_provider = '',
    string $dns_challenge_resolver = ''
  ) {
    $this->name = $name;
    $this->domain = $domain;
    $this->http_port = $http_port;
    $this->https_port = $https_port;
    $this->cert_filename = $cert_filename;
    $this->key_filename = $key_filename;
    $this->env = $env;
    $this->tls = $tls;
    $this->dns_challenge = $dns_challenge;
    $this->dns_challenge_provider = $dns_challenge_provider;
    $this->dns_challenge_resolver = $dns_challenge_resolver;
  }

  /**
   * @param bool $addon_portainer
   */
  public function setAddonPortainer(bool $addon_portainer): void {
    $this->addon_portainer = $addon_portainer;
  }

  /**
   * @param string $hub_token
   */
  public function setHubToken(string $hub_token): void {
    $this->hub_token = $hub_token;
  }

  /**
   * Update the Traefik container.
   */
  public function update(): void {
    // Update host wide traefik container.
    $traefikPath = $_SERVER['HOME'] . '/.traefik';
    $traefikCertPath = $_SERVER['HOME'] . '/.traefik/certs';
    $traefikConfigPath = $_SERVER['HOME'] . '/.traefik/configuration';
    $traefikFile = $traefikPath . '/docker-compose.yml';

    $user = getenv('USER');
    $fs = new Filesystem();
    if (!$fs->exists($traefikPath)) {
      $fs->mkdir($traefikPath);
      if ($user) {
        $fs->chown($traefikPath, $user, TRUE);
      }
    }
    if (!$fs->exists($traefikCertPath)) {
      $fs->mkdir($traefikCertPath);
      if ($user) {
        $fs->chown($traefikCertPath, $user, TRUE);
      }
    }
    if (!$fs->exists($traefikConfigPath)) {
      $fs->mkdir($traefikConfigPath);
      if ($user) {
        $fs->chown($traefikConfigPath, $user, TRUE);
      }
    }
    $upd1 = $this->updateFile($traefikConfigPath . '/certificates.toml', $this->defaultCertificatesConfig());
    $upd2 = $this->updateFile($traefikFile, Yaml::dump($this->defaultDockerCompose(), 9, 2));
    if ($upd1 || $upd2) {
      if ($user) {
        if (file_exists($traefikConfigPath . '/certificates.toml')) {
          $fs->chown($traefikConfigPath . '/certificates.toml', $user, TRUE);
        }
        if (file_exists($traefikFile)) {
          $fs->chown($traefikFile, $user, TRUE);
        }
      }
      $cwd = getcwd();
      chdir($traefikPath);
      exec('docker network create traefik-public 2>/dev/null');
      exec('docker compose --project-name traefik up -d');
      chdir($cwd);
    }
  }

  /**
   * @param string $filename
   * @param string $content
   *
   * @return bool
   */
  private function updateFile(string $filename, string $content): bool {
    if ($content === '') {
      if (file_exists($filename)) {
        unlink($filename);
        return TRUE;
      }
      return FALSE;
    }
    if (file_exists($filename) && file_get_contents($filename) === $content) {
      return FALSE;
    }
    file_put_contents($filename, $content);
    return TRUE;
  }

  /**
   * @return array
   */
  private function defaultDockerCompose(): array {
    $config = [
      'version' => '3.3',
      'services' => [
        'traefik' => [
          'container_name' => 'traefik',
          'image' => 'traefik:2.11',
          'command' => [
            '--api=true',
            '--api.dashboard=true',
            '--api.insecure=true',
            '--log.level=warning',
            '--entrypoints.web.address=:80',
            '--providers.file.directory=/configuration/',
            '--providers.file.watch=true',
            '--providers.docker=true',
            '--providers.docker.exposedbydefault=false',
          ],
          'restart' => 'unless-stopped',
          'networks' => [
            'traefik-public',
          ],
          'ports' => [
            $this->http_port . ':80',
          ],
          'labels' => [
            'traefik.enable=true',
            'traefik.network=traefik-public',
            'traefik.http.routers.traefik.service=api@internal',
            'traefik.http.routers.traefik.rule=Host(`traefik.' . $this->domain . '`)',
          ],
          'volumes' => [
            './certs:/certs/:ro',
            './configuration:/configuration/:ro',
            '/var/run/docker.sock:/var/run/docker.sock:ro',
          ],
        ],
      ],
      'networks' => [
        'traefik-public' => [
          'external' => true,
        ],
      ],
    ];
    if (!empty($this->env)) {
      $config['services']['traefik']['environment'] = $this->env;
    }
    if ($this->addon_portainer) {
      $config['services']['portainer'] = [
        'container_name' => 'portainer',
        'image' => 'portainer/portainer-ce:2.17.1',
        'command' => '-H unix:///var/run/docker.sock',
        'restart' => 'unless-stopped',
        'networks' => [
          'traefik-public',
        ],
        'labels' => [
          'traefik.enable=true',
          'traefik.network=traefik-public',
          'traefik.http.routers.frontend.rule=Host(`portainer.' . $this->domain . '`)',
          'traefik.http.routers.frontend.entrypoints=websecure',
          'traefik.http.services.frontend.loadbalancer.server.port=9000',
          'traefik.http.routers.frontend.service=frontend',
          'traefik.http.routers.edge.rule=Host(`edge.' . $this->domain . '`)',
          'traefik.http.routers.edge.entrypoints=websecure',
          'traefik.http.services.edge.loadbalancer.server.port=8000',
          'traefik.http.routers.edge.service=edge',
        ],
        'volumes' => [
          './portainerdata:/data',
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
      ];
    }
    if ($this->hub_token) {
      $config['services']['traefik']['command'][] = '--experimental.hub=true';
      $config['services']['traefik']['command'][] = '--hub.tls.insecure=true';
      $config['services']['traefik']['command'][] = '--metrics.prometheus.addrouterslabels=true';
      $config['services']['traefik']['ports'][] = '9900:9900';
      $config['services']['traefik']['ports'][] = '9901:9901';
      $config['services']['hub-agent'] = [
        'container_name' => 'hub-agent',
        'image' => 'ghcr.io/traefik/hub-agent-traefik:v1.0.1',
        'command' => [
          'run',
          '--hub.token=' . $this->hub_token,
          '--auth-server.advertise-url=http://hub-agent',
          '--traefik.host=traefik',
          '--traefik.tls.insecure=true',
        ],
        'restart' => 'on-failure',
        'networks' => [
          'traefik-public',
        ],
        'volumes' => [
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
        'depends_on' => [
          'traefik',
        ],
      ];
    }
    if ($this->tls) {
      $config['services']['traefik']['command'][] = '--entrypoints.websecure.address=:443';
      $config['services']['traefik']['command'][] = '--entrypoints.web.http.redirections.entrypoint.to=websecure';
      $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.email=admin@' . $this->domain;
      $config['services']['traefik']['ports'][] = $this->https_port . ':443';
      $config['services']['traefik']['labels'][] = 'traefik.http.routers.traefik.tls=true';
      $config['services']['traefik']['labels'][] = 'traefik.http.routers.traefik.tls.certresolver=lakedrops';
      if ($this->addon_portainer) {
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.frontend.tls=true';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.frontend.tls.certresolver=lakedrops';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.edge.tls=true';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.edge.tls.certresolver=lakedrops';
      }
      if ($this->dns_challenge && $this->dns_challenge_provider !== '') {
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge=true';
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge.provider=' . $this->dns_challenge_provider;
        if ($this->dns_challenge_resolver !== '') {
          $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge.resolvers=' . $this->dns_challenge_resolver;
        }
      }
      else {
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.storage=/etc/traefik/acme/acme.json';
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.httpchallenge.entrypoint=web';
      }
    }
    return $config;
  }

  /**
   * @return string
   */
  private function defaultCertificatesConfig(): string {
    if ($this->cert_filename === '' || $this->key_filename === '') {
      return '';
    }
    return <<<EOF
[[tls.certificates]]
   certFile = "/certs/$this->cert_filename"
   keyFile = "/certs/$this->key_filename"
EOF;

  }

}
